# Ignoring some linting rules in tests
# pylint: disable=redefined-outer-name
# pylint: disable=missing-docstring
import numpy as np

from bingo.evolutionary_algorithms.age_fitness import AgeFitnessEA
from bingo.evaluation.evaluation import Evaluation
from bingo.evolutionary_optimizers.island import Island
from bingo.local_optimizers.scipy_optimizer import ScipyOptimizer
from bingo.local_optimizers.local_opt_fitness import LocalOptFitnessFunction
from bingo.symbolic_regression import (
    ComponentGenerator,
    AGraphGenerator,
    AGraphCrossover,
    AGraphMutation,
    ExplicitRegression,
    ExplicitTrainingData,
)

from bingo.util.log import configure_logging

configure_logging(diagnostics_file="diagnostics.csv")

POP_SIZE = 128
STACK_SIZE = 10
MUTATION_PROBABILITY = 0.4
CROSSOVER_PROBABILITY = 0.4
NUM_POINTS = 100
START = -10
STOP = 10
ERROR_TOLERANCE = 1e-6


def init_x_vals(start, stop, num_points):
    return np.linspace(start, stop, num_points).reshape([-1, 1])


def equation_eval(x):
    return x ** 2 + 3.5 * x ** 3


def init_island():
    np.random.seed(4)
    x = init_x_vals(START, STOP, NUM_POINTS)
    y = equation_eval(x)
    training_data = ExplicitTrainingData(x, y)

    component_generator = ComponentGenerator(x.shape[1])
    component_generator.add_operator("+")
    component_generator.add_operator("-")
    component_generator.add_operator("*")

    crossover = AGraphCrossover()
    mutation = AGraphMutation(component_generator)

    agraph_generator = AGraphGenerator(STACK_SIZE, component_generator)

    fitness = ExplicitRegression(training_data=training_data)
    optimizer = ScipyOptimizer(fitness, method="lm")
    local_opt_fitness = LocalOptFitnessFunction(fitness, optimizer)
    evaluator = Evaluation(local_opt_fitness)

    ea = AgeFitnessEA(
        evaluator,
        agraph_generator,
        crossover,
        mutation,
        MUTATION_PROBABILITY,
        CROSSOVER_PROBABILITY,
        POP_SIZE,
    )

    island = Island(ea, agraph_generator, POP_SIZE)
    return island


def main():
    test_island = init_island()
    report_island_status(test_island)
    test_island.evolve_until_convergence(
        max_generations=1000, fitness_threshold=ERROR_TOLERANCE
    )
    report_island_status(test_island)


def report_island_status(test_island):
    print("-----  Generation %d  -----" % test_island.generational_age)
    print("Best individual:     ", test_island.get_best_individual())
    print("Best fitness:        ", test_island.get_best_fitness())
    print("Fitness evaluations: ", test_island.get_fitness_evaluation_count())


if __name__ == "__main__":
    main()
